//Pathways implemented here: PN->KC<->GGN->LH; PN->LH; PN->GGN

#include <stdio.h>
#include <stdlib.h> 
#include <math.h> 
#include <map>

//config and commandline parameters
#include "params.h"

using namespace std;

//---------Define the parameters---------------
//Parameters for full model (Nitin)
//input has to be changed with nPN change!
//#define   nPN    830	// number of PNs
//#define   nKC    50000  // number of KC cells
//#define   nLH    100    // number of nLH
//set nLH to 0 if you want to obtain pure PN->KC<->GGN model


//---------------------------------------------
//Values from Assisi 07, just 1/3 model
#define   nKC    15000  // number of KC cells
#define   nGGN   1      // single GGN neuron
#define   nPN    300    // number of PNs
//#define   nLH    100    // number of nLH; this value is not correctly rescaled to 1/3 model
//----------------------------------------------

#define   nLH   40    // rescale for smaller model


//-- Regular spiking neuron cell---------------------------------------
class RS{
  double y, xp, xpp, sigma_n, beta_n;	//Bazh 05, eq.1; xp==xn,xpp==x_(n-1),y==yn
  double sigma_dc, beta_dc;
public:
  double alpha, x;				//Bazh 05, eq.2; x==f()
  double sigma, sigma_e, beta_e, Idc, S_CX_DEND, mu;
  int spike;					//spike triggered

  RS(){ init(); }

  //Mapping parameters to particular papers.
  //Let those params fixed, instead create your own method.
  void params_Bazhenov05(){
    mu=0.0005;		//Bazh 05
    alpha=3.65; 	//Bazh 05
    sigma=0.06;		//Bazh 05
    sigma_e=1.0;	//Bazh 05
    beta_e=0.133;
  }
  void params_Assisi(){ 
    params_Bazhenov05();

    //experiments with additional KC variability
    sigma_dc=1.0;	//temporary from Collins, add variant input to KC
    beta_dc=0.133;	//temporary from Collins, add variant input to KC; beta_e from B05
    Idc = 0.00;		//constant direct current
  }
  void init(){
    spike=0;
    //parameters
    params_Assisi();
    beta_e=0.03;	
    S_CX_DEND = 165.0e-6; //area of the cell 

    //initial conditions
    xpp=-1+sigma;
    xp=-1+sigma;
    x=-1+sigma;
    y= x - alpha/(1-x);


  }

  inline void calc(double);
};

void RS::calc(double I){
  //synaptic input currents
  beta_n = beta_e * I + beta_dc * Idc;
  sigma_n = sigma_e * I + sigma_dc * Idc;

  //limit total synaptic input Bazh 05 (-0.0001;0.1) 
  //     if (beta_n < -0.0001) beta_n = -0.0001;
  if (beta_n < -1.0) beta_n = -1.0;
  if (beta_n > 1.0) beta_n = 1.0;

  //Bazh 05, eq.2
  if(xp <= 0.0) {
    x = alpha / (1.0 - xp) + y +beta_n;
    spike = 0;
  }
  else{
    if(xp <= alpha + y +beta_n && xpp <= 0.0) {
      x = alpha + y + beta_n;
      spike = 1;
    }
    else {
      x = -1;
      spike = 0;
    }
  }
  //Bazh 05, eq.1
  y = y - mu* (xp +1.0) + mu * sigma + mu * sigma_n;
  xpp = xp;
  xp = x;
}



//--Non-Spiking Neuron Cell---------------------------------------
class NS{
public:
  double x, y;		//main variables
  double mu, alpha, sigma;  //main parameters
  //not in official equations
  double sigma_e;	//shaping input
  double S_CX_DEND;	//area of cell

  NS(){ init(); }

  //basic set of parameters, do not touch.	
  //instead create your own params_method like in RS.
  void init(){
    //parameters, GGN map model #2, mail from Rulkov
    mu=0.005;
    alpha=0.8;
    sigma=-0.5;

    sigma_e=1.0;		//just in case we need shape input
    S_CX_DEND=165.0e-6;		//blindly inherited from RS;

    //initial conditions
    x=-1+sigma;			//fixed point, baseline
    y= alpha*(x-x*x*x/27)-x;	//fixed point  
  }

  inline void calc(double);
};

//GGN map model #2
void NS::calc(double I){
  double xn=x;
  double sigma_n = sigma_e * I;	

  if((-3 <= xn) && (xn <= 3))  x = xn - xn*xn*xn/27;
   else if (xn < -3) x = -2; 
    else if (xn > 3) x =  2; 

  x = alpha * x - y; 
  y = y + mu * (1+xn) - mu * (sigma + sigma_n);

}



//------------------------------------------------------------------
class AMPAmap1 {
  static const double E_AMPA = -0.0; //reversal potential
  static const double gamma  =  0.4; //conductance decay

public:
  double I, g;	//Current from synapse to soma; Syn. conductance
  AMPAmap1() {
    I=0;
    g=0;
  }
  inline void calc(double, double, int);
};
void AMPAmap1::calc(double g_AMPA, double x_post, int spike) {

  g = gamma * g;
  if(spike) g += g_AMPA;

  //Without saturation GGN has oscillatory artifacts due to large input
  double v = x_post - E_AMPA;
  if (v > 0) v = 0;

  I = - g * v;
}

//--GABA_A inhibition--------------------------------------------------
class GABAAmap1 {
 static const double E_GABAA = -1.1;
 static const double gamma   =  0.3;

public:
 double I,g;
 GABAAmap1() {
     I=0;
     g=0;
 }
 inline void calc(double, double, double);
};
void GABAAmap1::calc(double g_GABA, double x_post, double x) { // x used to be spike

      double value, result, ti;
      //reponse to input impulse (spike)
      if(x > -1.4){ 	//threshold for activity;  
        value = (-x - (-1.5)) / 1.5;
	result = exp (value);
	ti = 1 / (1 + result);
        g = gamma * g + (ti * g_GABA); //gamma * g + spike;
      }
      //afterspike decrease
      else{
        g = gamma * g;
      }
      I = - g * (x_post - E_GABAA);
} 
//-----------------------------------------------------------------

//Input from PN neurons
//The file is usually too big for [neuron][time] array
//so we just load file and feed the computation sequentially with increasing time
class PNInput {
  multimap<int,int> spikes;	//time, neuron number;

 public:
  void loadInput(){	//load from file
    FILE *f;
    int time, neuron;
    if (!(f=fopen("spikesPN.txt","r"))) { fprintf(stderr,"spikesPN doesn't exist\n"); exit(1); }

    while (fscanf(f, "%d %d", &time, &neuron) == 2)
      spikes.insert(pair<int, int>(time, neuron)); //sort by time
   
   fclose(f);
  }//loadInput

 void getTimeSliceSpikes(int t, int spikesPN[]){	//prepare input PN spikes for the current time t
   for (int i=0; i<nPN; i++) spikesPN[i]=0;

   //O(log N), can be improved to O(1) if we build vector and remember last position
   pair<multimap<int, int>::iterator, multimap<int, int>::iterator> r;
   r = spikes.equal_range(t);  //get spikes for the current time

   for (multimap<int,int>::iterator it = r.first; it!=r.second; it++)
     spikesPN[(*it).second]=1;
 }
};//PNInput

int num_threads=4;	//default number of threads, can by changed by commandline par. -num_threads X or parameters file

int   iters;           //number of iterations

//------Connectivity (bool)---------
int      C_KCPN[nKC][nPN];  
int      C_GGNKC[nGGN][nKC];
int      C_KCGGN[nKC][nGGN];
int      C_LHGGN[nLH][nGGN];
int      C_GGNPN[nGGN][nPN]; 
int      C_LHPN[nLH][nPN]; 
//------Duplicate connectivity stored as vector for speedup.
//------(we have enough memory but not enough time.)
//V[][0] is number of edges, V[i][j] = edge V[i][j]->i
int      V_KCPN[nKC][nPN+1];
int      V_GGNKC[nGGN][nKC+1];
int      V_KCGGN[nKC][nGGN+1];
int      V_LHGGN[nLH][nGGN+1];
int      V_GGNPN[nGGN][nPN+1];
int      V_LHPN[nLH][nPN+1];


//--------Synapses-----------------
AMPAmap1           ampa_KCPN[nKC][nPN];
AMPAmap1           ampa_GGNKC[nGGN][nKC];
GABAAmap1          gaba_KCGGN[nKC][nGGN];
GABAAmap1          gaba_LHGGN[nLH][nGGN];
AMPAmap1           ampa_GGNPN[nGGN][nPN];
AMPAmap1           ampa_LHPN[nLH][nPN];

//---Synaptic strength (external params)-----
double   g_ampa_GGNKC; //AMPA from KC to GGN
double   g_ampa_KCPN;  //AMPA from PN to KC
double   g_gaba_KCGGN; //GABA_A from GGN to KC
double   g_gaba_LHGGN; //GABA_A from GGN to LH
double   g_ampa_GGNPN; //AMPA from PN to GGN
double   g_ampa_LHPN;  //AMPA from PN to LH

//-------Neurons-----------------
RS            KC[nKC];   // KC cells
NS            GGN[nGGN]; // GGN      
RS            LH[nLH];   // Lateral horn interneurons        

int      spikesPN[nPN]; //input spikes from PN in the current timeslice

double   R; //normalized random number

FILE     *fspikesPN;
FILE     *fspikesKC;
FILE     *fspikesGGN;
FILE     *fspikesLH;
FILE     *fISynGGN;
FILE     *finterKC;
FILE	 *fKCSig;
FILE	 *fKCPN;
FILE     *f1, *fweights, *fmeanweight;

//-----------function to evaluate the Rhs of all the maps---
void fun(int t){

  int i, j, k, e;
  double IsynKCPN[nKC];	//summary of all synaptic input for given KC/scaled by number of connections 
  double IsynKCGGN[nKC];
  double IsynLHGGN[nLH];
  double IsynGGNKC[nGGN];
  double IsynGGNPN[nGGN];
  double IsynLHPN[nLH];

  //------------ calculating synaptic currents-----------------------------------------

//We optimalize for openMP the first loop only, most of computation happens here
#pragma omp parallel for shared(IsynKCPN,V_KCPN,ampa_KCPN,KC,spikesPN,g_ampa_KCPN) private(i,j,k,e) schedule(guided,10) num_threads(num_threads)
  // calculating AMPA type current from PN to KC-------------------------------------
  for(i = 0; i < nKC; i++){
    double tmp =  0.0;
    k = V_KCPN[i][0];
    for(j = 1; j <= k; j++){
	e = V_KCPN[i][j];  //e->i
	ampa_KCPN[i][e].calc(g_ampa_KCPN/KC[i].S_CX_DEND, KC[i].x, spikesPN[e]);
	tmp+= ampa_KCPN[i][e].I/k;
    }//for j
    IsynKCPN[i]=tmp;
  }//for i 


  // calculating AMPA type current from KC to GGN-------------------------------------
  for(i = 0; i < nGGN; i++){
    IsynGGNKC[i] = 0.0;
    k = V_GGNKC[i][0];
    for(j = 1; j <= k; j++){
	e = V_GGNKC[i][j];  //e->i
	ampa_GGNKC[i][e].calc(g_ampa_GGNKC/GGN[i].S_CX_DEND, GGN[i].x, KC[e].spike);
	IsynGGNKC[i] += ampa_GGNKC[i][e].I/k;
    } //for j
  }//for i 


  // calculating GABA_A type current from GGN to KC-------------------------------------
  for(i = 0; i < nKC; i++){
    IsynKCGGN[i] = 0.0;
    k = V_KCGGN[i][0];
    for(j = 1; j <= k; j++){
	e = V_KCGGN[i][j];  //e->i
	gaba_KCGGN[i][e].calc(g_gaba_KCGGN/KC[i].S_CX_DEND, KC[i].x, GGN[e].x);
	IsynKCGGN[i] += gaba_KCGGN[i][e].I/k;
    }//for j
  } //for i


  // calculating GABA_A type current from GGN to LH-------------------------------------
  for(i = 0; i < nLH; i++){
    IsynLHGGN[i] = 0.0;
    k = V_LHGGN[i][0];
    for(j = 1; j <= k; j++){
      e = V_LHGGN[i][j];  //e->i
      gaba_LHGGN[i][e].calc(g_gaba_LHGGN/LH[i].S_CX_DEND, LH[i].x, GGN[e].x);
      IsynLHGGN[i] += gaba_LHGGN[i][e].I/k;
    }//for j
  } //for i


  // calculating AMPA type current from PN to GGN-------------------------------------
  for(i = 0; i < nGGN; i++){
    IsynGGNPN[i] = 0.0;
    k = V_GGNPN[i][0];
    for(j = 1; j <= k; j++){
      e = V_GGNPN[i][j];  //e->i
      ampa_GGNPN[i][e].calc(g_ampa_GGNPN/GGN[i].S_CX_DEND, GGN[i].x, spikesPN[e]);
      IsynGGNPN[i] += ampa_GGNPN[i][e].I/k;
    }//for j
  }//for i 


  // calculating AMPA type current from PN to LH-------------------------------------
  for(i = 0; i < nLH; i++){
    IsynLHPN[i] = 0.0;
    k = V_LHPN[i][0];
    for(j = 1; j <= k; j++){
      e = V_LHPN[i][j];  //e->i
      ampa_LHPN[i][e].calc(g_ampa_LHPN/LH[i].S_CX_DEND, LH[i].x, spikesPN[e]);
      IsynLHPN[i] += ampa_LHPN[i][e].I/k;
    }//for j
  }//for i 


  //----calculate the state of all KC at the next iteration based on synaptic input-------
  for(i = 0; i < nKC; i++)
    KC[i].calc(IsynKCPN[i] + IsynKCGGN[i]);      

  //----calculate the state of GGN at the next iteration based on synaptic input-------
  for(i = 0; i < nGGN; i++){
    GGN[i].calc(IsynGGNKC[i] + IsynGGNPN[i]);      
    fprintf(fISynGGN,"%lf \t %lf \n",IsynGGNKC[i], IsynGGNPN[i]);
  }
  //-----calculate the state of all LH interneurons------------------------------
  for(i = 0; i < nLH; i++)
    LH[i].calc(IsynLHPN[i] + IsynLHGGN[i]);      
 
} //end fun

//create more efficient representation from bool connectivity matrix
void duplicate_connectivity(){
  int i,j;

  for (i=0; i<nKC; i++)
    for (j=0; j<nGGN; j++)
      if (C_KCGGN[i][j]) V_KCGGN[i][++V_KCGGN[i][0]]=j;

  for (i=0; i<nKC; i++)
    for (j=0; j<nPN; j++)
      if (C_KCPN[i][j]) V_KCPN[i][++V_KCPN[i][0]]=j;

  for (i=0; i<nGGN; i++)
    for (j=0; j<nKC; j++)
      if (C_GGNKC[i][j]) V_GGNKC[i][++V_GGNKC[i][0]]=j;

  for (i=0; i<nLH; i++)
    for (j=0; j<nGGN; j++)
      if (C_LHGGN[i][j]) V_LHGGN[i][++V_LHGGN[i][0]]=j;

  for (i=0; i<nGGN; i++)
    for (j=0; j<nPN; j++)
      if (C_GGNPN[i][j]) V_GGNPN[i][++V_GGNPN[i][0]]=j;

  for (i=0; i<nLH; i++)
    for (j=0; j<nPN; j++)
      if (C_LHPN[i][j]) V_LHPN[i][++V_LHPN[i][0]]=j;

}//duplicate connectivity

int main(int argc, char **argv)
{

  int  t,ii,jj,xx,yy,i,j;
  double kk,kcsum;

  double kc_mu = 0.004, kc_mu_distrib = 0.0005; //temporary external parameters for playing with mu
  double lh_mu = 0.0005, lh_mu_distrib = 0.000; //temporary external parameters for playing with mu
  PNInput PNInputSpikes;
  PNInputSpikes.loadInput();

  //***************initialization********//
  //---initialize connectivity matrix-----

  //get external parameters for the model
  add_int_param(iters);
  add_double_param(g_ampa_GGNKC);
  add_double_param(g_ampa_KCPN);
  add_double_param(g_gaba_KCGGN);
  add_double_param(g_ampa_GGNPN);
  add_double_param(g_ampa_LHPN);
  add_double_param(g_gaba_LHGGN);
  add_double_param(kc_mu);
  add_double_param(kc_mu_distrib);
  add_double_param(lh_mu);
  add_double_param(lh_mu_distrib);
  add_int_param(num_threads);
  assert(load_parameters(argv[1]));
  assert(cmdline_parameters(argc,argv));
  print_parameters();


  //-----read the connectivity matrix-----------------
  //randomly generated

  //TODO centralize usage of (s)rand
//  srand(1);
//  for(i=0; i<nKC; i++)
//    for(j=0; j<nPN; j++){
//      R = 1.0 * rand()/(RAND_MAX + 1.0);
      
//      if(R < 0.5)
//	C_KCPN[i][j]=1;  // P(PN to KC) = 0.5
//    }

 fKCPN = fopen("KCPN_5000_rand.txt","r");
 while(fscanf(fKCPN,"%d %d", &xx, &yy)==2){
   ii = xx;
   jj = yy;
   C_KCPN[ii][jj] = 1;
 }
 fclose(fKCPN);

  srand(2);
  for(i=0; i<nGGN; i++)
    for(j=0; j<nKC; j++){
      R = 1.0 * rand()/(RAND_MAX + 1.0);
      
      if(R < 2)
	C_GGNKC[i][j]=1;  // all KCs connected to all GGNs
    }

  srand(3);
  for(i=0; i<nKC; i++)
    for(j=0; j<nGGN; j++){
      R = 1.0 * rand()/(RAND_MAX + 1.0);
      
      if(R < 2)
	C_KCGGN[i][j]=1;  // all GGNs connected to all KCs 
    }

  srand(5);
  for(i=0; i<nLH; i++)
    for(j=0; j<nGGN; j++){
      R = 1.0 * rand()/(RAND_MAX + 1.0);
      
      if(R < 2)
	C_LHGGN[i][j]=1;  // all GGNs connected to all LHs 
    }
   
  //TODO kill this
  srand(4);
  for(i=0; i<nGGN; i++)
    for(j=0; j<nPN; j++){
      R = 1.0 * rand()/(RAND_MAX + 1.0);
      
      if(R < 2)
	C_GGNPN[i][j]=1;  // P(PN to GGN) = 1
    }

   srand(4);
   for(i=0; i<nLH; i++)
     for(j=0; j<nPN; j++){
       R = 1.0 * rand()/(RAND_MAX + 1.0);
       if(R < 0.7)
         C_LHPN[i][j]=1;  // P(PN to GGN) = 0.7
     }

  duplicate_connectivity();

  //-----end read the connectivity matrix-----------------


  //**************end initialization*****************//

  //srand(2);
  //1. variability in KCs
  //for(ii = 0; ii < nKC; ++ii){
    //R = 2.0 * rand()/(RAND_MAX + 1.0) - 1.0;
    //KC[ii].sigma = KC[ii].sigma + R*0.0005;

  //}

//1. read KC.sigma from KCSig.txt
  fKCSig = fopen("KCSig3.txt","r");
  for(ii = 0; ii < nKC; ++ii){
    if (fscanf(fKCSig,"%lf",&kk)!=1) {fprintf(stderr,"KCSig3 input error.\n");return 1;};
    KC[ii].sigma = kk;
  }
  fclose(fKCSig);

  //srand(2);
  //1. variability in KCs
  //for(ii = 0; ii < nKC; ++ii){
    //R = 2.0 * rand()/(RAND_MAX + 1.0) - 1.0;
   // KC[ii].sigma = KC[ii].sigma + R*0.04;

  //}

  srand(7);
  //1. variability in KC mu
  for(ii = 0; ii < nKC; ++ii){
  temp:
    R = 2.0 * rand()/(RAND_MAX + 1.0) - 1.0;
    KC[ii].mu = kc_mu + R*kc_mu_distrib;
    if (KC[ii].mu<=0) goto temp;
  }

  srand(7);
  //1. variability in LH mu
  for(ii = 0; ii < nLH; ++ii){
    R = 2.0 * rand()/(RAND_MAX + 1.0) - 1.0;
    LH[ii].mu = lh_mu + R*lh_mu_distrib;
  }

//leftover from previous experiments where all KCs got additional input
//  srand(2);
//  3. DC input to KCs with variability
//  for(ii = 0; ii < nKC; ++ii) {
//    KC[ii].Idc = 0.012 * rand()/RAND_MAX;
//}

  //----------- end KC variability----------------


  fspikesKC = fopen("spikesKC","w");
  finterKC = fopen("interKC","w"); 
  fspikesGGN = fopen("spikesGGN","w");
  fISynGGN = fopen("SynGGN","w");
  fspikesLH = fopen("spikesLH","w");

 //-----------Start iteration --------------------
  t = 0;
  while (t < iters){

    if (t % 100 == 0)
      fprintf(stderr,"%d \n",t);

    //read PN input spike vector
    PNInputSpikes.getTimeSliceSpikes(t,spikesPN);

    //evaluate the rhs of all diffeqns
    fun(t);

    //write KC spikes to file
    for (kcsum = ii = 0; ii < nKC; ++ii){
      if(KC[ii].spike == 1){
        fprintf(fspikesKC,"%d %d \n", t, ii);
      }
      kcsum += KC[i].x;
    }

    fprintf(finterKC,"%d %f %f %f %f \n", t, kcsum/nKC, KC[490].x, KC[801].x, KC[561].x);
    
    //write GGN spikes to file
    for (ii = 0; ii < nGGN; ++ii)
      fprintf(fspikesGGN,"%lf \t",GGN[ii].x);
    fprintf(fspikesGGN,"\n");

    for (ii = 0; ii < nLH; ++ii)
      if(LH[ii].spike == 1)
	fprintf(fspikesLH,"%d %d \n", t, ii);

    ++t;
  }

  fclose(fspikesKC);
  fclose(finterKC);
  fclose(fspikesGGN);
  fclose(fISynGGN);  
  fclose(fspikesLH);
  return 0;
} //end main

